function [ sent_score, sent_type, lang, response ] = get_alchemy_sentiment(text, nattempts, delay_sec, min_nwords, min_nchars, use_cache, cache_fpath)
%GET_ALCHEMY_SENTIMENT Quantifies sentiment of text through AlchemyAPI's (http://www.alchemyapi.com) API.
%
%  NOTE:
%
%		In order to use AlchemyAPI's web-service, you need to request an API Key from
%		AlchemyAPI and, then, assign it in the accompanying get_alchemy_apikey.m file.
%
%  INPUT:
%
%		text -- the text to extract sentiment from
%
%		nattempts -- number of attempts to contact the web-service; default is 3
%
%		delay_sec -- delay (# of sec) between the web-service contact attempts; default is 5(sec)
%
%		min_nwords -- minimal number of words in text for sentiment analysis; default is 3. If the actual
%					  number of words is less than min_nwords, then the web-service is not accessed, and
%					  sent_score is assigned nan.
%
%		min_nchars -- same as min_nwords, but for the minimal number of characters in text; default is 15
%
%		use_cache -- if true (which it is by default), then each request to AlchemyAPI is cached is a
%					 locally stored .mat file, so that the sentiment for the already processed texts will
%					 be extracted from the cache, rather than from the web-service. The cost of reading
%					 from cache is nominal, as the .mat is loaded only once. The cost of updating the
%					 cache may, sometimes, be substantial, as the updated hashtable is flushed to the .mat
%					 file after /each/ new text is analyzed. On the bright side, you need to quantify the
%					 sentiment of your texts only once, and, then, all requests to GET_ALCHEMY_SENTIMENT
%					 will be processed through cache. Regardless of whether you decide to cache through
%					 a .mat file, the function always caches its results in RAM via a persistent text2sent
%					 hashtable.
%
%		cache_fpath -- the path to the .mat file serving as cache; default is './alchemy_sentiment_bank.mat'.
%
%  OUTPUT:
%
%		sent_score -- a real value between -1 and +1, quantifying the text's sentiment. May be nan if it
%					is impossible to quantify sentiment (-- cannot reach the web-service, cannot understand
%					the text, the text is too short to analyze, etc.)
%
%		sent_type -- type of sentiment (positive, negative, neutral, unknown). Typically, sent_score and
%					 its type (isnan?) and/or sign (+/-) is enough.
%
%		lang -- the detected language of text
%
%		response -- the raw response from the web-service
%
%  EXAMPLE:
%
%	>> text = 'Finally, let understand that when we stand together, we will always win. When men and women stand together for justice, we win. When black, white and Hispanic people stand together for justice, we win.';
%	>> sent = get_alchemy_sentiment(text) % with caching to/from a .mat file
%
%		sent = 0.7724
%
%	>> use_cache = false;
%	>> text = 'We have to have a wall. We have to have a border. And in that wall we?re going to have a big fat door where people can come into the country, but they have to come in legally.';
%	>> [sent, sent_type, lang] = get_alchemy_sentiment(text, nan, nan, nan, nan, use_cache)
%
%		sent = -0.2848
%		sent_type = negative
%		lang = english
%
%	N.B. Do not forget to check sent for NaN!
%

if(nargin < 1 || ~ischar(text))
	error('Text is not provided or invalid.');
end
text = strtrim(text);
if(nargin < 2 || isnan(nattempts))
	nattempts = 3;
end
if(nargin < 3 || isnan(delay_sec))
	delay_sec = 5;
end
if(nargin < 4 || isnan(min_nwords))
	min_nwords = 3;
end
if(nargin < 5 || isnan(min_nchars))
	min_nchars = 15;
end
if(nargin < 6 || ~islogical(use_cache) && ~isnumeric(use_cache))
	use_cache = true;
end
if(nargin < 7)
	alchemy_basepath = fileparts(mfilename('fullpath'));
	sent_bank_fpath = fullfile(alchemy_basepath, './alchemy_sentiment_bank.mat');
else
	sent_bank_fpath = cache_fpath;
end

response = '';

persistent text2sent;
if(length(text) >= min_nchars && length(str2words(text)) >= min_nwords)
	
	if(isempty(text2sent))
		if(use_cache && exist(sent_bank_fpath, 'file') == 2)
			text2sent = load(sent_bank_fpath);
			text2sent = text2sent.text2sent;
		else
			text2sent = containers.Map;
		end
	end
	if(text2sent.isKey(text))
		sent = text2sent(text);
		unstruct(sent);
		return;
	end

	gateway = 'https://gateway-a.watsonplatform.net/calls/text/TextGetTextSentiment';
	apikey = get_alchemy_apikey();
	outputMode = 'xml';

	itrial = 1;
	response_received = false;
	while(~response_received && itrial <= nattempts)
		try
			response = webread(gateway, 'apikey', apikey, 'outputMode', outputMode, 'text', text);
		catch err
			warning('[Trial %d/%d] Unable to quantify sentiment of "%s". Reason: %s. Sleeping for %dsec...', itrial, nattempts, text, err.message, delay_sec);
			pause(delay_sec);
			itrial = itrial + 1;
			continue;
		end
		response_received = true;
		response_has_error = ~isempty(strfind(response, '<status>ERROR</status>'));
	end

	if(~isempty(response))
		
		response_fields = [];
		if(~response_has_error)
			pattern_with_mixed_score    = '<results>\s*?<status>\s*?(?<STATUS>[^<]*?)\s*?<\/status>\s*?(?:<usage>[^<]+<\/usage>)?\s*?<totalTransactions>\s*?\d+\s*?<\/totalTransactions>\s*?<language>\s*?(?<LANG>[^<]+)\s*?<\/language>\s*?<docSentiment>\s*?<mixed>\s*?[^<]+\s*?<\/mixed>\s*?<score>\s*?(?<SENTSCORE>[+-\.\d]+)\s*?<\/score>\s*?<type>\s*?(?<SENTTYPE>[^<]+)\s*?<\/type>\s*?<\/docSentiment>\s*?<\/results>';
			pattern_with_score          = '<results>\s*?<status>\s*?(?<STATUS>[^<]*?)\s*?<\/status>\s*?(?:<usage>[^<]+<\/usage>)?\s*?<totalTransactions>\s*?\d+\s*?<\/totalTransactions>\s*?<language>\s*?(?<LANG>[^<]+)\s*?<\/language>\s*?<docSentiment>\s*?                                 <score>\s*?(?<SENTSCORE>[+-\.\d]+)\s*?<\/score>\s*?<type>\s*?(?<SENTTYPE>[^<]+)\s*?<\/type>\s*?<\/docSentiment>\s*?<\/results>';
			pattern_without_score       = '<results>\s*?<status>\s*?(?<STATUS>[^<]*?)\s*?<\/status>\s*?(?:<usage>[^<]+<\/usage>)?\s*?<totalTransactions>\s*?\d+\s*?<\/totalTransactions>\s*?<language>\s*?(?<LANG>[^<]+)\s*?<\/language>\s*?<docSentiment>\s*?                                                                                    <type>\s*?(?<SENTTYPE>[^<]+)\s*?<\/type>\s*?<\/docSentiment>\s*?<\/results>';

			response_fields = regexp(response, pattern_with_mixed_score, 'names', 'freespacing', 'once');
			if(isempty(response_fields))
				response_fields = regexp(response, pattern_with_score, 'names', 'freespacing', 'once');
				if(isempty(response_fields))
					response_fields = regexp(response, pattern_without_score, 'names', 'freespacing', 'once');
				end
			end
		end

		if(~isempty(response_fields) || response_has_error)
			status = '';
			if(~response_has_error)
				status = response_fields.STATUS;
			end
			if(strcmp('OK', status))
				sent_type = response_fields.SENTTYPE;
				if(~strcmp('neutral', sent_type) || isfield(response_fields, 'SENTSCORE'))
					sent_score = str2double(response_fields.SENTSCORE);
				else
					sent_score = 0;
				end
				lang = response_fields.LANG;
			elseif(~isempty(strfind(response, '<status>ERROR</status>')))
				sent_type = 'ERROR';
				sent_score = 0;
				lang = 'ERROR';
			end

			text2sent(text) = struct( ...
				'sent_score', sent_score, ...
				'sent_type', sent_type, ...
				'lang', lang, ...
				'response', response ...
			);
			if(use_cache)
				save(sent_bank_fpath, 'text2sent', '-v7');
			end

			return;
		end
	end
end

sent_score = nan;
sent_type = 'unknown';
lang = 'unknown';

end


